@@ -4,10 +4,10 @@
 #
 # Copyright (C) 2007 Oliver Charles
 # Copyright (C) 2007, 2010-2011 Lukáš Lalinský
-# Copyright (C) 2007-2011, 2014, 2018-2021 Philipp Wolfer
+# Copyright (C) 2007-2011, 2014, 2018-2022 Philipp Wolfer
 # Copyright (C) 2011 Michael Wiencek
 # Copyright (C) 2011-2012, 2015 Wieland Hoffmann
-# Copyright (C) 2013-2015, 2018-2019 Laurent Monin
+# Copyright (C) 2013-2015, 2018-2021 Laurent Monin
 # Copyright (C) 2016 Ville Skyttä
 # Copyright (C) 2016-2018 Sambhav Kothari
 # Copyright (C) 2017 Antonio Larrosa
@@ -43,19 +43,24 @@ from PyQt5.QtCore import (
 from picard import log
 from picard.config import get_config
 from picard.const import DEFAULT_COVER_IMAGE_FILENAME
-from picard.coverart.utils import translate_caa_type
+from picard.coverart.utils import (
+    Id3ImageType,
+    image_type_as_id3_num,
+    translate_caa_type,
+)
 from picard.metadata import Metadata
 from picard.util import (
     decode_filename,
     encode_filename,
     imageinfo,
     is_absolute_path,
+    periodictouch,
 )
 from picard.util.scripttofilename import script_to_filename
 
 
 _datafiles = dict()
-_datafile_mutex = QMutex(QMutex.Recursive)
+_datafile_mutex = QMutex(QMutex.RecursionMode.Recursive)
 
 
 class DataHash:
@@ -64,7 +69,7 @@ class DataHash:
         self._filename = None
         _datafile_mutex.lock()
         try:
-            m = md5()
+            m = md5()  # nosec
             m.update(data)
             self._hash = m.hexdigest()
             if self._hash not in _datafiles:
@@ -73,6 +78,7 @@ class DataHash:
                 with os.fdopen(fd, "wb") as imagefile:
                     imagefile.write(data)
                 _datafiles[self._hash] = self._filename
+                periodictouch.register_file(self._filename)
                 log.debug("Saving image data %s to %r" % (self._hash, self._filename))
             else:
                 self._filename = _datafiles[self._hash]
@@ -89,6 +95,7 @@ class DataHash:
         if self._filename:
             try:
                 os.unlink(self._filename)
+                periodictouch.unregister_file(self._filename)
             except BaseException:
                 pass
             else:
@@ -137,7 +144,8 @@ class CoverArtImage:
     is_front = None
     sourceprefix = "URL"
 
-    def __init__(self, url=None, types=None, comment='', data=None, support_types=None, support_multi_types=None):
+    def __init__(self, url=None, types=None, comment='', data=None, support_types=None,
+                 support_multi_types=None, id3_type=None):
         if types is None:
             self.types = []
         else:
@@ -153,6 +161,7 @@ class CoverArtImage:
         self.can_be_saved_to_tags = True
         self.can_be_saved_to_disk = True
         self.can_be_saved_to_metadata = True
+        self.id3_type = id3_type
         if support_types is not None:
             self.support_types = support_types
         if support_multi_types is not None:
@@ -164,7 +173,7 @@ class CoverArtImage:
         self.url = QUrl(url)
         self.host = self.url.host()
         self.port = self.url.port(443 if self.url.scheme() == 'https' else 80)
-        self.path = self.url.path(QUrl.FullyEncoded)
+        self.path = self.url.path(QUrl.ComponentFormattingOption.FullyEncoded)
         if self.url.hasQuery():
             query = QUrlQuery(self.url.query())
             self.queryargs = dict(query.queryItems())
@@ -238,10 +247,7 @@ class CoverArtImage:
                     return (self.datahash, self.maintype) == (other.datahash, other.maintype)
             else:
                 return self.datahash == other.datahash
-        elif not self and not other:
-            return True
-        else:
-            return False
+        return not self and not other
 
     def __hash__(self):
         if self.datahash is None:
@@ -279,6 +285,24 @@ class CoverArtImage:
         # TODO: do something better than randomly using the first in the list
         return self.types[0]
 
+    @property
+    def id3_type(self):
+        """Returns the ID3 APIC image type.
+        If explicitly set the type is returned as is, otherwise it is derived
+        from `maintype`. See http://www.id3.org/id3v2.4.0-frames"""
+        if self._id3_type is not None:
+            return self._id3_type
+        else:
+            return image_type_as_id3_num(self.maintype)
+
+    @id3_type.setter
+    def id3_type(self, type):
+        """Explicitly sets the ID3 APIC image type.
+        If set to None the type will be derived from `maintype`."""
+        if type is not None:
+            type = Id3ImageType(type)
+        self._id3_type = type
+
     def _make_image_filename(self, filename, dirname, _metadata):
         metadata = Metadata()
         metadata.copy(_metadata)
@@ -337,7 +361,8 @@ class CoverArtImage:
             except OSError as e:
                 raise CoverArtImageIOError(e)
 
-    def _next_filename(self, filename, counters):
+    @staticmethod
+    def _next_filename(filename, counters):
         if counters[filename]:
             new_filename = "%s (%d)" % (decode_filename(filename), counters[filename])
         else:
@@ -414,8 +439,8 @@ class TagCoverArtImage(CoverArtImage):
 
     def __init__(self, file, tag=None, types=None, is_front=None,
                  support_types=False, comment='', data=None,
-                 support_multi_types=False):
-        super().__init__(url=None, types=types, comment=comment, data=data)
+                 support_multi_types=False, id3_type=None):
+        super().__init__(url=None, types=types, comment=comment, data=data, id3_type=id3_type)
         self.sourcefile = file
         self.tag = tag
         self.support_types = support_types
