@@ -3,11 +3,11 @@
 # Picard, the next-generation MusicBrainz tagger
 #
 # Copyright (C) 2006-2007, 2011 Lukáš Lalinský
-# Copyright (C) 2009-2011, 2014, 2018-2020 Philipp Wolfer
+# Copyright (C) 2009-2011, 2014, 2018-2021 Philipp Wolfer
 # Copyright (C) 2011-2014 Wieland Hoffmann
 # Copyright (C) 2012-2013 Michael Wiencek
 # Copyright (C) 2013 Calvin Walton
-# Copyright (C) 2013-2014, 2018-2019 Laurent Monin
+# Copyright (C) 2013-2014, 2018-2021 Laurent Monin
 # Copyright (C) 2014-2015, 2017 Sophist-UK
 # Copyright (C) 2016-2018 Sambhav Kothari
 #
@@ -39,11 +39,8 @@ from picard.coverart.image import (
     CoverArtImageError,
     TagCoverArtImage,
 )
+from picard.coverart.utils import types_from_id3
 from picard.file import File
-from picard.formats.id3 import (
-    image_type_as_id3_num,
-    types_from_id3,
-)
 from picard.formats.mutagenext import delall_ci
 from picard.metadata import Metadata
 from picard.util import encode_filename
@@ -188,7 +185,7 @@ class ASFFile(File):
         'work': 'WM/Work',
         'website': 'WM/AuthorURL',
     }
-    __RTRANS = dict([(b, a) for a, b in __TRANS.items()])
+    __RTRANS = {b: a for a, b in __TRANS.items()}
 
     # Tags to load case insensitive
     __TRANS_CI = {
@@ -200,7 +197,7 @@ class ASFFile(File):
         'replaygain_track_range': 'REPLAYGAIN_TRACK_RANGE',
         'replaygain_reference_loudness': 'REPLAYGAIN_REFERENCE_LOUDNESS',
     }
-    __RTRANS_CI = dict([(b.lower(), a) for a, b in __TRANS_CI.items()])
+    __RTRANS_CI = {b.lower(): a for a, b in __TRANS_CI.items()}
 
     def __init__(self, filename):
         super().__init__(filename)
@@ -216,7 +213,7 @@ class ASFFile(File):
             if name == 'WM/Picture':
                 for image in values:
                     try:
-                        (mime, data, type_, description) = unpack_image(image.value)
+                        (mime, data, image_type, description) = unpack_image(image.value)
                     except ValueError as e:
                         log.warning('Cannot unpack image from %r: %s',
                                     filename, e)
@@ -225,10 +222,11 @@ class ASFFile(File):
                         coverartimage = TagCoverArtImage(
                             file=filename,
                             tag=name,
-                            types=types_from_id3(type_),
+                            types=types_from_id3(image_type),
                             comment=description,
                             support_types=True,
                             data=data,
+                            id3_type=image_type,
                         )
                     except CoverArtImageError as e:
                         log.error('Cannot load image from %r: %s' %
@@ -267,11 +265,13 @@ class ASFFile(File):
         tags = file.tags
 
         if config.setting['clear_existing_tags']:
+            cover = tags.get('WM/Picture') if config.setting['preserve_images'] else None
             tags.clear()
+            if cover:
+                tags['WM/Picture'] = cover
         cover = []
         for image in metadata.images.to_be_saved_to_tags():
-            tag_data = pack_image(image.mimetype, image.data,
-                                  image_type_as_id3_num(image.maintype),
+            tag_data = pack_image(image.mimetype, image.data, image.id3_type,
                                   image.comment)
             cover.append(ASFByteArrayAttribute(tag_data))
         if cover:
@@ -310,7 +310,7 @@ class ASFFile(File):
     def supports_tag(cls, name):
         return (name in cls.__TRANS
                 or name in cls.__TRANS_CI
-                or name in ('~rating', '~length', 'totaldiscs')
+                or name in {'~rating', 'totaldiscs'}
                 or name.startswith('lyrics:'))
 
     def _get_tag_name(self, name):
