@@ -4,7 +4,7 @@
 #
 # Copyright (C) 2004 Robert Kaye
 # Copyright (C) 2006-2009, 2011-2012, 2014 Lukáš Lalinský
-# Copyright (C) 2008-2011, 2014, 2018-2020 Philipp Wolfer
+# Copyright (C) 2008-2011, 2014, 2018-2022 Philipp Wolfer
 # Copyright (C) 2009 Carlin Mangar
 # Copyright (C) 2009 david
 # Copyright (C) 2010 fatih
@@ -12,15 +12,17 @@
 # Copyright (C) 2012, 2014-2015 Wieland Hoffmann
 # Copyright (C) 2013 Ionuț Ciocîrlan
 # Copyright (C) 2013-2014 Sophist-UK
-# Copyright (C) 2013-2014, 2018-2020 Laurent Monin
+# Copyright (C) 2013-2014, 2018-2021 Laurent Monin
 # Copyright (C) 2014 Johannes Dewender
 # Copyright (C) 2016 Rahul Raturi
 # Copyright (C) 2016 barami
 # Copyright (C) 2016-2018 Sambhav Kothari
 # Copyright (C) 2017 Frederik “Freso” S. Olesen
-# Copyright (C) 2018 Bob Swift
 # Copyright (C) 2018 Vishal Choudhary
+# Copyright (C) 2018, 2021 Bob Swift
 # Copyright (C) 2020 Ray Bouchard
+# Copyright (C) 2021 Gabriel Ferreira
+# Copyright (C) 2021 Louis Sautier
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU General Public License
@@ -40,13 +42,15 @@
 import builtins
 from collections import namedtuple
 from collections.abc import Mapping
-import html
 from itertools import chain
 import json
+from locale import strxfrm as _strxfrm
 import ntpath
 from operator import attrgetter
 import os
+from pathlib import PurePath
 import re
+import subprocess  # nosec: B404
 import sys
 from time import monotonic
 import unicodedata
@@ -56,16 +60,32 @@ from dateutil.parser import parse
 from PyQt5 import QtCore
 
 from picard import log
-from picard.const import MUSICBRAINZ_SERVERS
+from picard.const import (
+    DEFAULT_COPY_TEXT,
+    DEFAULT_NUMBERED_TITLE_FORMAT,
+    MUSICBRAINZ_SERVERS,
+)
 from picard.const.sys import (
     FROZEN_TEMP_PATH,
     IS_FROZEN,
+    IS_MACOS,
     IS_WIN,
 )
 
 
 if IS_WIN:
-    from ctypes import windll
+    import winreg
+
+# Windows path length constraints
+# See https://docs.microsoft.com/en-us/windows/win32/fileio/maximum-file-path-limitation
+# the entire path's length (260 - 1 null character)
+WIN_MAX_FILEPATH_LEN = 259
+# the entire parent directory path's length must leave room for a 8.3 filename
+WIN_MAX_DIRPATH_LEN = WIN_MAX_FILEPATH_LEN - 12
+# a single node's (directory or file) length
+WIN_MAX_NODE_LEN = 255
+# Prefix for long paths in Windows API
+WIN_LONGPATH_PREFIX = '\\\\?\\'
 
 
 class LockableObject(QtCore.QObject):
@@ -160,13 +180,58 @@ def decode_filename(filename):
         return filename.decode(_io_encoding)
 
 
+def _check_windows_min_version(major, build):
+    try:
+        v = sys.getwindowsversion()
+        return v.major >= major and v.build >= build
+    except AttributeError:
+        return False
+
+
+def system_supports_long_paths():
+    """Detects long path support.
+
+    On Windows returns True, only if long path support is enabled in the registry (Windows 10 1607 or later).
+    All other systems return always True.
+    """
+    if not IS_WIN:
+        return True
+    try:
+        # Use cached value
+        return system_supports_long_paths._supported
+    except AttributeError:
+        pass
+    try:
+        # Long path support can be enabled in Windows 10 version 1607 or later
+        if _check_windows_min_version(10, 14393):
+            with winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE,
+                                r"SYSTEM\CurrentControlSet\Control\FileSystem") as key:
+                supported = winreg.QueryValueEx(key, "LongPathsEnabled")[0] == 1
+        else:
+            supported = False
+        system_supports_long_paths._supported = supported
+        return supported
+    except OSError:
+        log.info('Failed reading LongPathsEnabled from registry')
+        return False
+
+
 def normpath(path):
+    path = os.path.normpath(path)
     try:
         path = os.path.realpath(path)
     except OSError as why:
         # realpath can fail if path does not exist or is not accessible
+        # or on Windows if drives are mounted without mount manager
+        # (see https://tickets.metabrainz.org/browse/PICARD-2425).
         log.warning('Failed getting realpath for "%s": %s', path, why)
-    return os.path.normpath(path)
+    # If the path is longer than 259 characters on Windows, prepend the \\?\
+    # prefix. This enables access to long paths using the Windows API. See
+    # https://docs.microsoft.com/en-us/windows/win32/fileio/maximum-file-path-limitation
+    if (IS_WIN and len(path) > WIN_MAX_FILEPATH_LEN and not system_supports_long_paths()
+        and not path.startswith(WIN_LONGPATH_PREFIX)):
+        path = WIN_LONGPATH_PREFIX + path
+    return path
 
 
 def is_absolute_path(path):
@@ -251,6 +316,17 @@ def sanitize_filename(string, repl="_", win_compat=False):
     return string
 
 
+def make_filename_from_title(title=None, default=None):
+    if default is None:
+        default = _("No Title")
+    if not title or not title.strip():
+        title = default
+    filename = sanitize_filename(title, win_compat=IS_WIN)
+    if IS_WIN:
+        filename = replace_win32_incompat(filename)
+    return filename
+
+
 def _reverse_sortname(sortname):
     """Reverse sortnames."""
     chunks = [a.strip() for a in sortname.split(",")]
@@ -290,11 +366,17 @@ def find_existing_path(path):
     return decode_filename(path)
 
 
+def _add_windows_executable_extension(*executables):
+    return [e if e.endswith(('.py', '.exe')) else e + '.exe' for e in executables]
+
+
 def find_executable(*executables):
     if IS_WIN:
-        executables = [e + '.exe' for e in executables]
+        executables = _add_windows_executable_extension(*executables)
     paths = [os.path.dirname(sys.executable)] if sys.executable else []
     paths += os.environ.get('PATH', '').split(os.pathsep)
+    paths.append('./')
+
     # This is for searching for executables bundled in packaged builds
     if IS_FROZEN:
         paths += [FROZEN_TEMP_PATH]
@@ -302,7 +384,34 @@ def find_executable(*executables):
         for executable in executables:
             f = os.path.join(path, executable)
             if os.path.isfile(f):
-                return f
+                return os.path.abspath(f)
+
+
+def run_executable(executable, *args, timeout=None):
+    # Prevent new shell window from appearing
+    startupinfo = None
+    if IS_WIN:
+        startupinfo = subprocess.STARTUPINFO()
+        startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
+
+    # Include python interpreter if running a python script
+    if ".py" in executable:
+        arguments = [sys.executable, executable, *args]
+    else:
+        arguments = [executable, *args]
+
+    # Call program with arguments
+    ret = subprocess.run(  # nosec: B603
+        arguments,
+        stdin=subprocess.DEVNULL,
+        stdout=subprocess.PIPE,
+        stderr=subprocess.PIPE,
+        startupinfo=startupinfo,
+        timeout=timeout
+    )
+
+    # Return (error code, stdout and stderr)
+    return ret.returncode, ret.stdout.decode(sys.stdout.encoding), ret.stderr.decode(sys.stderr.encoding)
 
 
 _mbid_format = '[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$'
@@ -366,6 +475,38 @@ def throttle(interval):
     return decorator
 
 
+class IgnoreUpdatesContext:
+    """Context manager for holding a boolean value, indicating whether updates are performed or not.
+    By default the context resolves to False. If entered it is True. This allows
+    to temporarily set a state on a block of code like:
+
+        ignore_changes = IgnoreUpdatesContext()
+        # Initially ignore_changes is False
+        with ignore_changes:
+            # Perform some tasks with ignore_changes now being True
+            ...
+        # ignore_changes is False again
+
+    The code actually doing updates can check `ignore_changes` and only perform
+    updates if it is `False`.
+    """
+
+    def __init__(self, onexit=None):
+        self._entered = 0
+        self._onexit = onexit
+
+    def __enter__(self):
+        self._entered += 1
+
+    def __exit__(self, type, value, tb):
+        self._entered -= 1
+        if self._onexit:
+            self._onexit()
+
+    def __bool__(self):
+        return self._entered > 0
+
+
 def uniqify(seq):
     """Uniqify a list, preserving order"""
     return list(iter_unique(seq))
@@ -378,37 +519,61 @@ def iter_unique(seq):
 
 
 # order is important
-_tracknum_regexps = (
+_tracknum_regexps = [re.compile(r, re.I) for r in (
     # search for explicit track number (prefix "track")
-    r"track[\s_-]*(?:no|nr)?[\s_-]*(\d+)",
-    # search for 2-digit number at start of string
-    r"^(\d{2})\D",
-    # search for 2-digit number at end of string
-    r"\D(\d{2})$",
-)
+    r"track[\s_-]*(?:(?:no|nr)\.?)?[\s_-]*(?P<number>\d+)",
+    # search for 1- or 2-digit number at start of string (additional leading zeroes are allowed)
+    # An optional disc number preceding the track number is ignored.
+    r"^(?:\d+[\s_-])?(?P<number>0*\d{1,2})(?:\.)[^0-9,]",  # "99. ", but not "99.02"
+    r"^(?:\d+[\s_-])?(?P<number>0*\d{1,2})[^0-9,.s]",
+    # search for 2-digit number at end of string (additional leading zeroes are allowed)
+    r"[^0-9,.\w](?P<number>0*\d{2})$",
+    r"[^0-9,.\w]\[(?P<number>0*\d{1,2})\]$",
+    r"[^0-9,.\w]\((?P<number>0*\d{2})\)$",
+    # File names which consist of only a number
+    r"^(?P<number>\d+)$",
+)]
 
 
 def tracknum_from_filename(base_filename):
     """Guess and extract track number from filename
     Returns `None` if none found, the number as integer else
     """
-    filename, _ = os.path.splitext(base_filename)
-    for r in _tracknum_regexps:
-        match = re.search(r, filename, re.I)
+    filename, _ext = os.path.splitext(base_filename)
+    for pattern in _tracknum_regexps:
+        match = pattern.search(filename)
         if match:
-            n = int(match.group(1))
-            if n > 0:
+            n = int(match.group('number'))
+            # Numbers above 1900 are often years, track numbers should be much
+            # smaller even for extensive collections
+            if n > 0 and n < 1900:
                 return n
-    # find all numbers between 1 and 99
-    # 4-digit or more numbers are very unlikely to be a track number
-    # smaller number is preferred in any case
-    numbers = sorted([int(n) for n in re.findall(r'\d+', filename) if
-                      0 < int(n) <= 99])
-    if numbers:
-        return numbers[0]
     return None
 
 
+GuessedFromFilename = namedtuple('GuessedFromFilename', ('tracknumber', 'title'))
+
+
+def tracknum_and_title_from_filename(base_filename):
+    """Guess tracknumber and title from filename.
+    Uses `tracknum_from_filename` to guess the tracknumber. The filename is used
+    as the title. If the tracknumber is at the beginning of the title it gets stripped.
+
+    Returns a tuple `(tracknumber, title)`.
+    """
+    filename, _ext = os.path.splitext(base_filename)
+    title = filename
+    tracknumber = tracknum_from_filename(base_filename)
+    if tracknumber is not None:
+        tracknumber = str(tracknumber)
+        stripped_filename = filename.lstrip('0')
+        tnlen = len(tracknumber)
+        if stripped_filename[:tnlen] == tracknumber:
+            title = stripped_filename[tnlen:].lstrip()
+
+    return GuessedFromFilename(tracknumber, title)
+
+
 def is_hidden(filepath):
     """Test whether a file or directory is hidden.
     A file is considered hidden if it starts with a dot
@@ -419,17 +584,29 @@ def is_hidden(filepath):
         or _has_hidden_attribute(filepath)
 
 
-def _has_hidden_attribute(filepath):
-    if not IS_WIN:
-        return False
-    # FIXME: On OSX detecting hidden files involves more
-    # than just checking for dot files, see
-    # https://stackoverflow.com/questions/284115/cross-platform-hidden-file-detection
-    try:
-        attrs = windll.kernel32.GetFileAttributesW(filepath)
-        assert attrs != -1
-        return bool(attrs & 2)
-    except (AttributeError, AssertionError):
+if IS_WIN:
+    from ctypes import windll
+
+    def _has_hidden_attribute(filepath):
+        try:
+            attrs = windll.kernel32.GetFileAttributesW(filepath)
+            assert attrs != -1
+            return bool(attrs & 2)
+        except (AttributeError, AssertionError):
+            return False
+
+elif IS_MACOS:
+    import Foundation
+
+    def _has_hidden_attribute(filepath):
+        # On macOS detecting hidden files involves more than just checking for dot files, see
+        # https://stackoverflow.com/questions/284115/cross-platform-hidden-file-detection
+        url = Foundation.NSURL.fileURLWithPath_(filepath)
+        result = url.getResourceValue_forKey_error_(None, Foundation.NSURLIsHiddenKey, None)
+        return result[1]
+
+else:
+    def _has_hidden_attribute(filepath):
         return False
 
 
@@ -457,12 +634,21 @@ def linear_combination_of_weights(parts):
 
 
 def album_artist_from_path(filename, album, artist):
-    """If album is not set, try to extract album and artist from path
+    """If album is not set, try to extract album and artist from path.
+
+    Args:
+        filename: The full file path
+        album: Default album name
+        artist: Default artist name
+
+    Returns:
+        A tuple (album, artist)
     """
     if not album:
-        dirs = os.path.dirname(filename).replace('\\', '/').lstrip('/').split('/')
+        path = PurePath(filename)
+        dirs = list(path.relative_to(path.anchor).parent.parts)
         # Strip disc subdirectory from list
-        if re.search(r'\b(?:CD|DVD|Disc)\s*\d+\b', dirs[-1], re.I):
+        if dirs and re.search(r'\b(?:CD|DVD|Disc)\s*\d+\b', dirs[-1], re.I):
             del dirs[-1]
         if dirs:
             # For clustering assume %artist%/%album%/file or %artist% - %album%/file
@@ -486,13 +672,14 @@ def build_qurl(host, port=80, path=None, queryargs=None):
     """
     url = QtCore.QUrl()
     url.setHost(host)
-    url.setPort(port)
 
-    if host in MUSICBRAINZ_SERVERS or port == 443:
+    if port == 443 or host in MUSICBRAINZ_SERVERS:
         url.setScheme("https")
-        url.setPort(443)
+    elif port == 80:
+        url.setScheme("http")
     else:
         url.setScheme("http")
+        url.setPort(port)
 
     if path is not None:
         url.setPath(path)
@@ -560,10 +747,6 @@ def convert_to_string(obj):
 builtins.__dict__['string_'] = convert_to_string
 
 
-def htmlescape(string):
-    return html.escape(string, quote=False)
-
-
 def load_json(data):
     """Deserializes a string or bytes like json response and converts
     it to a python object.
@@ -625,24 +808,34 @@ def compare_barcodes(barcode1, barcode2):
     return barcode1.zfill(13) == barcode2.zfill(13)
 
 
-BestMatch = namedtuple('BestMatch', 'similarity result num_results')
+BestMatch = namedtuple('BestMatch', ('similarity', 'result'))
 
 
 def sort_by_similarity(candidates):
+    """Sorts the objects in candidates by similarity.
+
+    Args:
+        candidates: Iterable with objects having a `similarity`  attribute
+    Returns: List of candidates sorted by similarity (highest similarity first)
+    """
     return sorted(
-        candidates(),
+        candidates,
         reverse=True,
         key=attrgetter('similarity')
     )
 
 
 def find_best_match(candidates, no_match):
-    sorted_results = sort_by_similarity(candidates)
-    if sorted_results:
-        result = sorted_results[0]
-    else:
-        result = no_match
-    return BestMatch(similarity=result.similarity, result=result, num_results=len(sorted_results))
+    """Returns a BestMatch based on the similarity of candidates.
+
+    Args:
+        candidates: Iterable with objects having a `similarity`  attribute
+        no_match: Match to return if there was no candidate
+
+    Returns: `BestMatch` with the similarity and the matched object as result.
+    """
+    best_match = max(candidates, key=attrgetter('similarity'), default=no_match)
+    return BestMatch(similarity=best_match.similarity, result=best_match)
 
 
 def get_qt_enum(cls, enum):
@@ -708,3 +901,205 @@ def extract_year_from_date(dt):
             return parse(dt).year
     except (TypeError, ValueError):
         return None
+
+
+def pattern_as_regex(pattern, allow_wildcards=False, flags=0):
+    """Parses a string and interprets it as a matching pattern.
+
+    - If pattern is of the form /pattern/flags it is interpreted as a regular expression (e.g. `/foo.*/`).
+      The flags are optional and in addition to the flags passed in the `flags` function parameter. Supported
+      flags in the expression are "i" (ignore case) and "m" (multiline)
+    - Otherwise if `allow_wildcards` is True, it is interpreted as a pattern that allows wildcard matching (see below)
+    - If `allow_wildcards` is False a regex matching the literal string is returned
+
+    Wildcard matching currently supports these characters:
+    - `*`: Matches an arbitrary number of characters or none, e.g. `fo*` matches "foo" or "foot".
+    - `?`: Matches exactly one character, e.g. `fo?` matches "foo" or "for".
+    - `[...]`: Matches any character in the set, e.g. `[fo?]` matches all of "f", "o" and "?".
+    - `?`, `*`, `[`, `]` and `\\` can be escaped with a backslash \\ to match the literal
+      character, e.g. `fo\\?` matches "fo?".
+
+    Args:
+        pattern: The pattern as a string
+        allow_wildcards: If true and if the the pattern is not interpreted as a regex wildard matching is allowed.
+        flags: Additional regex flags to set (e.g. `re.I`)
+
+    Returns: An re.Pattern instance
+
+    Raises: `re.error` if the regular expression could not be parsed
+    """
+    plain_pattern = pattern.rstrip('im')
+    if len(plain_pattern) > 2 and plain_pattern[0] == '/' and plain_pattern[-1] == '/':
+        extra_flags = pattern[len(plain_pattern):]
+        if 'i' in extra_flags:
+            flags |= re.IGNORECASE
+        if 'm' in extra_flags:
+            flags |= re.MULTILINE
+        regex = plain_pattern[1:-1]
+    elif allow_wildcards:
+        regex = '^' + wildcards_to_regex_pattern(pattern) + '$'
+    else:
+        regex = re.escape(pattern)
+    return re.compile(regex, flags)
+
+
+def wildcards_to_regex_pattern(pattern):
+    """Converts a pattern with shell like wildcards into a regular expression string.
+
+    The following syntax is supported:
+    - `*`: Matches an arbitrary number of characters or none, e.g. `fo*` matches "foo" or "foot".
+    - `?`: Matches exactly one character, e.g. `fo?` matches "foo" or "for".
+    - `[...]`
+    - `?`, `*` and `\\` can be escaped with a backslash \\ to match the literal character, e.g. `fo\\?` matches "fo?".
+
+    Args:
+        pattern: The pattern as a string
+
+    Returns: A string with a valid regular expression.
+    """
+    regex = []
+    group = None
+    escape = False
+    for c in pattern:
+        if group is not None:
+            if escape:
+                if c in {'\\', '[', ']'}:
+                    c = '\\' + c
+                else:
+                    group.append('\\\\')
+                escape = False
+            if c == ']':
+                group.append(c)
+                part = ''.join(group)
+                group = None
+            elif c == '\\':
+                escape = True
+                continue
+            else:
+                group.append(c)
+                continue
+        elif escape:
+            if c in {'*', '?', '\\', '[', ']'}:
+                part = '\\' + c
+            else:
+                part = re.escape('\\' + c)
+            escape = False
+        elif c == '\\':
+            escape = True
+            continue
+        elif c == '[':
+            group = ['[']
+            continue
+        elif c == '*':
+            part = '.*'
+        elif c == '?':
+            part = '.'
+        else:
+            part = re.escape(c)
+        regex.append(part)
+
+    # There might be an unclosed character group. Interpret the starting
+    # bracket of the group as a literal bracket and re-evaluate the rest.
+    if group is not None:
+        regex.append('\\[')
+        regex.append(wildcards_to_regex_pattern(''.join(group[1:])))
+    return ''.join(regex)
+
+
+def _regex_numbered_title_fmt(fmt, title_repl, count_repl):
+    title_marker = '{title}'
+    count_marker = '{count}'
+
+    parts = fmt.split(title_marker)
+
+    def wrap_count(p):
+        if count_marker in p:
+            return '(?:' + re.escape(p) + ')?'
+        else:
+            return p
+
+    return (
+        re.escape(title_marker).join(wrap_count(p) for p in parts)
+        .replace(re.escape(title_marker), title_repl)
+        .replace(re.escape(count_marker), count_repl)
+    )
+
+
+def unique_numbered_title(default_title, existing_titles, fmt=None):
+    """Generate a new unique and numbered title
+       based on given default title and existing titles
+    """
+    if fmt is None:
+        fmt = _(DEFAULT_NUMBERED_TITLE_FORMAT)
+
+    escaped_title = re.escape(default_title)
+    reg_count = r'(\d+)'
+    regstr = _regex_numbered_title_fmt(fmt, escaped_title, reg_count)
+    regex = re.compile(regstr)
+    count = 0
+    for title in existing_titles:
+        m = regex.fullmatch(title)
+        if m:
+            num = m.group(1)
+            if num is not None:
+                count = max(count, int(num))
+            else:
+                count += 1
+    return fmt.format(title=default_title, count=count + 1)
+
+
+def get_base_title_with_suffix(title, suffix, fmt=None):
+    """Extract the base portion of a title,
+       removing the suffix and number portion from the end.
+    """
+    if fmt is None:
+        fmt = _(DEFAULT_NUMBERED_TITLE_FORMAT)
+
+    escaped_suffix = re.escape(suffix)
+    reg_title = r'(?P<title>.*?)(?:\s*' + escaped_suffix + ')?'
+    reg_count = r'\d*'
+    regstr = _regex_numbered_title_fmt(fmt, reg_title, reg_count)\
+        .replace(r'\ ', r'\s+')\
+        .replace(' ', r'\s+')
+    match_obj = re.fullmatch(regstr, title)
+    return match_obj['title'] if match_obj else title
+
+
+def get_base_title(title):
+    """Extract the base portion of a title, using the standard suffix.
+    """
+    suffix = _(DEFAULT_COPY_TEXT)
+    return get_base_title_with_suffix(title, suffix)
+
+
+def iter_exception_chain(err):
+    """Iterate over the exception chain.
+    Yields this exception and all __context__ and __cause__ exceptions"""
+    yield err
+    if hasattr(err, '__context__'):
+        yield from iter_exception_chain(err.__context__)
+    if hasattr(err, '__cause__'):
+        yield from iter_exception_chain(err.__cause__)
+
+
+def any_exception_isinstance(error, type_):
+    """Returns True, if any exception in the exception chain is instance of type_."""
+    return any(isinstance(err, type_) for err in iter_exception_chain(error))
+
+
+def strxfrm(string):
+    """Transforms a string to one that can be used in locale-aware comparisons.
+
+    Wrapper around locale.strxfrm, that never throws OSError. If an OSError
+    occurs this function will return the string converted to lower case.
+
+    Args:
+        string: The string to convert
+
+    Returns: A string that can be compared locale-aware
+    """
+    try:
+        return _strxfrm(string)
+    except OSError as err:
+        log.warning('strxfrm(%r) failed: %r', string, err)
+        return string.lower()
